package com.perforce.p4java.common.io;

import static com.perforce.p4java.common.quantity.Amount.DEFAULT_BUFFER_SIZE;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Arrays;

import javax.annotation.Nonnull;

/**
 * @author Sean Shou
 * @since 25/08/2016
 */
public class PdfFileDetector {
  private static final byte[] PDF_HEAD_MAGIC_NUMBER = {'%', 'P', 'D', 'F', '-'};

  /*
   * PDF tail magic number:
   * <pre>
   *      bytes                         comments
   * ------------------------          -----------------------------
   *  %%EOF SPACE OxOD 0x0A             win-style & before pdf_v1.3
   *  %%EOF SPACE 0x0A                  unix-style & before pdf_v1.3
   *  %%EOF OxOD  0x0A                  win-style line-ending
   *  %%EOF 0x0A                        unix-style line-ending
   *  %%EOF
   *
   * </pre>
   *
   */
  private static final byte[] PDF_MAX_TAIL_MAGIC_NUMBER = new byte[]{
      (byte) 0x25, // %
      (byte) 0x25, // %
      (byte) 0x45, // E
      (byte) 0x4F, // O
      (byte) 0x46, // F
      (byte) 0x20, // SPACE
      (byte) 0x0D, // EOL -- '\r'
      (byte) 0x0A  // EOL -- '\n'
  };

  private static final byte[] PDF_EOF_ONLY_TAIL_MAGIC_NUMBER = new byte[]{
      (byte) 0x25, // %
      (byte) 0x25, // %
      (byte) 0x45, // E
      (byte) 0x4F, // O
      (byte) 0x46  // F
  };

  private PdfFileDetector() { /* util */ }

  public static boolean isPdfFile(@Nonnull File file) throws IOException {
    return isPdfHeadMagicNumberPresent(file) && isPdfTailMagicNumberPresent(file);
  }

  /**
   * Return true IFF the contents seem to be a PDF.
   */
  private static boolean isPdfHeadMagicNumberPresent(@Nonnull File file) throws IOException {
    byte[] buffer = new byte[DEFAULT_BUFFER_SIZE];
    FileInputStream inputStream = new FileInputStream(file);
    boolean isPresent = false;
    try {
        int read = inputStream.read(buffer);
    	isPresent = read >= PDF_HEAD_MAGIC_NUMBER.length && isPdfHeadMagicNumberFound(buffer);
    } finally {
    	inputStream.close();
    }
    return isPresent;
  }

  private static boolean isPdfTailMagicNumberPresent(@Nonnull File file) throws IOException {
    int max = PDF_MAX_TAIL_MAGIC_NUMBER.length;
    byte[] tailNBytes = readTailNBytes(file, max);
    return isContainsDoublePercentEOFTailMagicNumber(tailNBytes);
  }

  private static byte[] readTailNBytes(@Nonnull File file, int n) throws IOException {
    byte[] buffer = new byte[n];
    try (RandomAccessFile randomAccessFile = new RandomAccessFile(file, "r")) {
      randomAccessFile.seek(file.length() - n);
      int read = randomAccessFile.read(buffer, 0, n);
      if (read > 0) {
        return Arrays.copyOf(buffer, read);
      } else {
        return new byte[0];
      }
    }
  }

  private static boolean isPdfHeadMagicNumberFound(byte[] sourceArray) {
    if (sourceArray.length < PdfFileDetector.PDF_HEAD_MAGIC_NUMBER.length) {
      return false;
    } else {
      byte[] headBytes = Arrays.copyOf(sourceArray, PdfFileDetector.PDF_HEAD_MAGIC_NUMBER.length);
      return Arrays.equals(headBytes, PdfFileDetector.PDF_HEAD_MAGIC_NUMBER);
    }
  }

  private static boolean isContainsDoublePercentEOFTailMagicNumber(byte[] sourceArray) {
    int length = PDF_EOF_ONLY_TAIL_MAGIC_NUMBER.length;
    if (sourceArray.length >= length) {
      int lastOffset = sourceArray.length - length;
      for (int i = 0; i <= lastOffset; ) {
        byte firstByteOfTailMagicNumber = PDF_EOF_ONLY_TAIL_MAGIC_NUMBER[0];
        if (sourceArray[i] == firstByteOfTailMagicNumber) {
          byte[] potentialMatchArray = Arrays.copyOfRange(sourceArray, i, i + length);
          if (Arrays.equals(potentialMatchArray, PDF_EOF_ONLY_TAIL_MAGIC_NUMBER)) {
            return true;
          }
        }
        i++;
      }
    }

    return false;
  }
}
