package com.perforce.p4java.common.quantity;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertTrue;

import org.junit.Test;

import com.google.common.collect.Lists;
import com.google.common.collect.Ordering;

/**
 * @author Sean
 */
public class AmountTest {

  @Test
  public void testEquals() {
    assertEquals("expected value equality semantics",
        Amount.of(1, Data.MB), Amount.of(1, Data.MB));

    assertEquals("expected equality to be calculated from amounts converted to a common unit",
        Amount.of(1L, Data.MB), Amount.of(1024L, Data.KB));

    assertNotEquals("expected unit conversions for equality tests to not lose precision",
        Amount.of(25L, Data.MB), Amount.of(1L, Data.GB));
    assertNotEquals("expected unit conversions for equality tests to not lose precision",
        Amount.of(1L, Data.GB), Amount.of(25L, Data.MB));

    assertFalse("expected value equality to work only for the same Number types",
        Amount.of(1L, Data.TB).equals(Amount.of(1.0, Data.TB)));
    assertFalse("expected value equality to work only for the same Number types",
        Amount.of(1L, Data.TB).equals(Amount.of(1, Data.TB)));

    assertFalse("amounts with incompatible units should never be equal even if their values are",
        Amount.of(1L, Data.MB).equals(Amount.of(1L, Data.BITS)));
  }

  @Test
  public void testComparisonMixedUnits() {
    assertTrue(Amount.of(1, Data.MB).compareTo(Amount.of(1023, Data.KB)) > 0);
    assertTrue(Amount.of(1, Data.MB).compareTo(Amount.of(1024, Data.KB)) == 0);
    assertTrue(Amount.of(1, Data.MB).compareTo(Amount.of(1025, Data.KB)) < 0);

    assertTrue(Amount.of(1023, Data.KB).compareTo(Amount.of(1, Data.MB)) < 0);
    assertTrue(Amount.of(1024, Data.KB).compareTo(Amount.of(1, Data.MB)) == 0);
    assertTrue(Amount.of(1025, Data.KB).compareTo(Amount.of(1, Data.MB)) > 0);
  }

  @Test
  @SuppressWarnings("unchecked") // Needed because type information lost in vargs.
  public void testOrderingMixedUnits() {
    assertEquals(
        Lists.newArrayList(
            Amount.of(1, Data.BITS),
            Amount.of(1, Data.KB),
            Amount.of(1, Data.MB),
            Amount.of(1, Data.MB)),
        Ordering.natural().sortedCopy(Lists.newArrayList(
            Amount.of(1, Data.KB),
            Amount.of(1024, Data.KB),
            Amount.of(1, Data.BITS),
            Amount.of(1, Data.MB))));
  }

  @Test
  @SuppressWarnings("unchecked") // Needed because type information lost in vargs.
  public void testOrderingSameUnits() {
    assertEquals(
        Lists.newArrayList(
            Amount.of(1, Data.BITS),
            Amount.of(2, Data.BITS),
            Amount.of(3, Data.BITS),
            Amount.of(4, Data.BITS)),
        Ordering.natural().sortedCopy(Lists.newArrayList(
            Amount.of(3, Data.BITS),
            Amount.of(2, Data.BITS),
            Amount.of(1, Data.BITS),
            Amount.of(4, Data.BITS))));
  }

  @Test
  public void testConvert() {
    Amount<Long, Data> fileSize = Amount.of(15L, Data.Mb);
    assertEquals(Long.valueOf(15 * 1024), fileSize.as(Data.Kb));

    assertEquals("expected conversion losing precision to use truncation", Long.valueOf(0), fileSize.as(Data.Gb));

    assertEquals("expected conversion losing precision to use truncation", Long.valueOf(0), Amount.of(45L, Data.Mb).as(Data.Gb));

    Amount<Double, Data> decimalFileSize = Amount.of(256.0, Data.Mb);
    assertEquals(Double.valueOf(256 * 1024), decimalFileSize.as(Data.Kb));
    assertEquals(Double.valueOf(0.25), decimalFileSize.as(Data.Gb));
  }

  @Test(expected = Amount.TypeOverflowException.class)
  public void testAmountThrowsTypeOverflowException() {
    Amount.of(1000, Time.DAYS).asChecked(Time.MILLISECONDS);
  }
}